!+
! Module: epreject_mod
!
! Authors:  Konrad Dennerl, Hermann Brunner, Vadim Burwitz - MPE
!          
! Description: 
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module epreject_mod

  use oal
  use cal
  use param
  use dal
  
  implicit none

  private

  public epreject

contains

  !---------------------------------------------------------------------
  subroutine epreject
  !---------------------------------------------------------------------
  
   real(kind=Single) :: hmed, expo, exp1k, h1, h2, bg, bgfac, offsig
   integer(kind=Int32) :: i, j, k, ix, iy, ky, jy, ixx, iyy, nevent, kval
   integer(kind=Int32) :: koff, ksel, iccd, ccd, idum, isubmode, flagged
   integer(kind=Int32) :: noffcorr, ios = 0, badX, noffsets, nbadpix
   integer(kind=Int32) :: winX0, winX1, winY0, winY1, win4Y0
   integer(kind=Int32) :: count3, pix3, count5, pix5
   integer(kind=Int16) :: i2off, i2pha
   integer(kind=Int32), parameter :: nxdim=64, nydim=200
   integer(kind=Int32), parameter :: nccd=12, minspec=16, maxspec=65, &
                                     maxoffcorr=500 
   integer(kind=Int32), dimension(nccd,nydim,maxspec) :: i4spc
   integer(kind=Int32), dimension(nydim,maxspec) :: i4buf, i4buff
   integer(kind=Int32), dimension(nxdim,nydim) :: i4pmap, i30map
   integer(kind=Int16), dimension(nxdim,nydim) :: i2pmap
   real(kind=Single), dimension(nydim,maxspec) :: spcfrc
   real(kind=Single), dimension(0:12) :: noiseparam
   integer(kind=Int32), dimension(nydim) :: i4med, indx 
   real(kind=Single), dimension(:), pointer :: hlo, hhi
   integer(kind=Int16), dimension(:), pointer :: val, rrawx
   real(kind=double), dimension(:), pointer :: timeCol
   integer(kind=Int16), dimension(:), pointer :: rawXCol, rawYCol, phaCol, offCol
   integer(kind=Int16), dimension(:), pointer :: noiseCol, flareCol, offXCol, bpRawXCol, bpRawYCol, yExtCol 
   integer(kind=Int8), dimension(:), pointer :: ccdCol     
   real(kind=Single), dimension(:,:), pointer :: r4med
   real(kind=Single), dimension(:,:,:), pointer :: noisemap
   logical, dimension(nxdim,nydim) :: mask = .false.
   logical :: exist
   character(len=1024) :: string
   character(len=32) :: submode
  
!-- eprej flare routine VB
   integer(kind=Int32), dimension(0:2000000) :: lc
   real(kind=single), dimension(0:2000000) :: lcs
   real(kind=double)   :: tbinsize,t0
   real(kind=double)   :: ss
   integer(kind=Int32) :: kmax,ndlc,sbox
   integer(kind=Int32) :: npharangeelements
 ! integer(kind=Int32) :: nsmoothelements
   real(kind=double),   dimension(0:2) :: smo
   integer(kind=Int32), dimension(0:1) :: thrpha
   real(kind=double) :: thrsig
    
   type(DataSetT) :: infile
   type(TableT)  :: events, offsets, badpix
   type(ColumnT) :: ccdnr, rawX, rawY, pha, offCor, noise, flare, offX
   type(ColumnT) :: bpRawX, bpRawY, yExtent 
   type(ColumnT) :: time

   i4spc = 0
   i4buf = 0
   i4buff = 0
   i4pmap = 0
   i30map = 0
   i2pmap = 0
   spcfrc = 0.


   offsig = realParameter("sigma")
   do i=0,12
     noiseparam(i) = realParameter("noiseparameters",i)
   end do

! open the event file
   infile = dataSet(stringParameter("eventset"),MODIFY)
   if (.not. hasBlock(infile,"EVENTS")) then 
     call error("noEvents","No EVENTS extension in input data set")
   endif

   events = table(infile,"EVENTS",MODIFY)
   nevent = numberOfRows(events)
  
   time = column(events,"TIME",READ)
   ccdnr = column(events,"CCDNR",READ)
   rawX = column(events,"RAWX",READ)
   rawY = column(events,"RAWY",READ)
   pha = column(events,"PHA",READ)
  
   timeCol     => real64Data(time)
   ccdCol      => int8Data(ccdnr)
   rawXCol     => int16Data(rawX)
   rawYCol     => int16Data(rawY)
   phaCol      => int16Data(pha)

   submode=stringAttribute(events,"SUBMODE")

   if (.not.booleanParameter("withsoftflarescreening")) then

! append OFF_COR column
   if (hasColumn(events,"OFF_COR")) then
     offCor = column(events,"OFF_COR",MODIFY)  
     offCol => int16Data(offCor)
   else
     offCor = addColumn (events, "OFF_COR", Integer16, "")
     offCol => int16Data(offCor)
     offCol = 0
   end if   

! append NOISE column
   if (hasColumn(events,"NOISE")) then
     noise = column(events,"NOISE",MODIFY)  
   else
     noise = addColumn (events, "NOISE", Integer16, "")
   end if   
   noiseCol => int16Data(noise)
   noiseCol = 0

   
   winX0 = int16Attribute(events,"WINDOWX0") 
   winY0 = int16Attribute(events,"WINDOWY0")     
   winX1 = int16Attribute(events,"WINDOWDX") + winX0 - 1
   winY1 = int16Attribute(events,"WINDOWDY") + winY0 - 1
   mask(winX0:winX1,winY0:winY1)=.true.

   expo=Real32Attribute(events,"EXPOSURE")
   exp1k=expo/1000.
!! expo =expo / 2.16 ! eFF

   if (index(submode,"PrimeFull")>0) then
     isubmode=PRIME_FULL_WINDOW
   else if (index(submode,"Ext")>0) then
     isubmode=PRIME_FULL_WINDOW_EXTENDED
   else if (index(submode,"Large")>0) then
     isubmode=PRIME_LARGE_WINDOW
   else if (index(submode,"Small")>0) then
     isubmode=PRIME_SMALL_WINDOW
   else
     isubmode=PRIME_FULL_WINDOW
     call warning("WrongSubMode","Could not read SUBMODE keyword; FF mode assumed") 
   end if

   iccd=ccdCol(0)
   call CAL_setState(instrument=EPN,ccdModeId=isubmode,ccdChipId=iccd)
   call CAL_pnOffsetCorrValues(hlo,hhi,val)
   noffcorr=size(val)
   call CAL_pnMedianMap(r4med)
   r4med=r4med*exp1k
   call CAL_pnNoiseMap(rrawx,noisemap)

! bin 20 & 30-40 ADU images    
   do i=0,nevent-1
     if (iccd /= ccdCol(i)) call error("Single CCD event file required") 
     if ((phaCol(i)+offCol(i))==20) then
       i4pmap(rawXCol(i),rawYCol(i))=i4pmap(rawXCol(i),rawYCol(i))+1
     end if
     if (phaCol(i)+offCol(i)>=30 .and. phaCol(i)+offCol(i)<=40) then
       i30map(rawXCol(i),rawYCol(i))=i30map(rawXCol(i),rawYCol(i))+1
     end if
   end do

! read in bad columns and remove from mask
   string = stringParameter("badcolumnset")   
   inquire(file=string,exist=exist)
   if (exist) then
     open(1,file=string,status="old")
     ios=0
     do while (ios == 0)
       read (1,*,iostat=ios) ccd, badX
       if (ios /= 0) exit
       if (ccd==iccd) mask(badX,winY0:winY1)=.false.
     end do       
     close(1)
   else
     call warning ("No bad column table found")
   end if

! get offset columns and remove from mask
   noffsets=0
   if (hasBlock(infile,"OFFSETS")) then
     offsets = table(infile,"OFFSETS",READ)
     noffsets = numberOfRows(offsets)
     
     if (noffsets > 0) then
       offX = column(offsets,"RAWX",READ)
       offXCol => int16Data(offX)

       do i=1,noffsets
         badX=offXCol(i-1)
         mask(badX,winY0:winY1)=.false.
       end do

       call release(offX)
     end if

     call release(offsets)   
   end if

! get badpixels and remove from mask
   nbadpix=0
   if (hasBlock(infile,"BADPIX")) then
     badpix = table(infile,"BADPIX",READ)
     nbadpix = numberOfRows(badpix)
   
     if (nbadpix > 0) then
       bpRawX = column(badpix,"RAWX",READ) 
       bpRawY = column(badpix,"RAWY",READ)
       yExtent = column(badpix,"YEXTENT",READ)

       bpRawXCol => int16Data(bpRawX)
       bpRawYCol => int16Data(bpRawY)
       yExtCol => int16Data(yExtent)
      
       do i=1,nbadpix
         do iy=bpRawYCol(i-1),bpRawYCol(i-1)+yExtCol(i-1)-1
           mask(bpRawXCol(i-1),iy)=.false.
         end do
       end do 

       call release(bpRawX)
       call release(bpRawY)
       call release(yExtent)
     end if

     call release(badpix)
   end if

! remove rows < 16 from mask
  mask(:,1:15)=.false.

! source search in 30-40 ADU map
   do ix=winX0,winX1
     do iy=winY0,winY1 
       count3=0
       pix3=0
       do ixx=ix-1,ix+1
         do iyy=iy-1,iy+1    
           if (ixx >= 1 .and. ixx <= nxdim .and. & 
             iyy >= 1 .and. iyy <= nydim) then
             if (mask(ixx,iyy)) then     
               count3=count3+i30map(ixx,iyy) 
               pix3=pix3+1
             end if
           end if   
         end do
       end do
       count5=0
       pix5=0 
       do ixx=ix-3,ix+3
         do iyy=iy-3,iy+3    
           if (ixx >= 1 .and. ixx <= nxdim .and. & 
             iyy >= 1 .and. iyy <= nydim) then
             if (mask(ixx,iyy)) then     
               count5=count5+i30map(ixx,iyy) 
               pix5=pix5+1
             end if
           end if   
         end do
       end do
       count5=count5-count3
       pix5=pix5-pix3
       if (pix3>0 .and. pix5>0) then
         bgfac=real(pix3)/real(pix5)
         bg=real(count5)*bgfac 
         if (count3-bg>10.*sqrt(bg)) then
           do ixx=ix-1,ix+1
             do iyy=iy-1,iy+1
               if (ixx >= 1 .and. ixx <= nxdim .and. & 
                   iyy >= 1 .and. iyy <= nydim) mask(ixx,iyy)=.false.
             end do
           end do  
         end if
       end if
     end do
   end do

   win4Y0 = (winY0-1)/4
   win4Y0 = 4*win4Y0+1

! search for additional hot pixels
!   do ix = winX0, winX1
!     do iy = win4Y0, winY1, 4
!       h4mean(1)=(i4pmap(ix,iy+1)+i4pmap(ix,iy+2)+i4pmap(ix,iy+3))/3.
!       h4mean(2)=(i4pmap(ix,iy)  +i4pmap(ix,iy+2)+i4pmap(ix,iy+3))/3.
!       h4mean(3)=(i4pmap(ix,iy)  +i4pmap(ix,iy+1)+i4pmap(ix,iy+3))/3.
!       h4mean(4)=(i4pmap(ix,iy)  +i4pmap(ix,iy+1)+i4pmap(ix,iy+2))/3.
!       do i=1,4
!         if (i4pmap(ix,iy+i-1)-h4mean(i)>20.*sqrt(h4mean(i))) mask(ix,iy+i-1)=.false.
!       end do
!     end do
!   end do 

   do ix = winX0, winX1
     do iy = win4Y0, winY1, 4

       kval=0
       hmed=0. 
       do i=0,3
         ky=iy+i
         if (mask(ix,ky)) then
           kval = kval + i4pmap(ix,ky) 
           hmed=hmed + r4med(ix-1,ky-1) 
         end if 
       end do

       if (abs(kval-hmed) > offsig*sqrt(real(kval))) then
         
         do ky=iy,iy+3 
         
           if (mask(ix,ky)) then
             kval = i4pmap(ix,ky) 
             hmed = r4med(ix-1,ky-1) 
       
             do i=0,noffcorr-1  
               if (kval-hmed > hlo(i)*exp1k .and. &
                   kval-hmed <= hhi(i)*exp1k) &
                   i2pmap(ix,ky) = val(i)
             end do
           end if
         end do 
       end if         
  
    end do
   end do

!   i2pmap(:,1:20) = 0
     
   koff = 0
   do i=0,nevent-1
     ix = rawXCol(i)
     iy = rawYCol(i)
     if (min(ix,iy).lt.1 .or. ix.gt.nxdim .or. iy.gt.nydim) then
       call warning ("Invalid RawX, RawY")
       offCol(i) = 0  
     else 
       i2off = i2pmap(ix,iy)
       i2pha = phaCol(i) + offCol(i) - i2off       
       offCol(i) = i2off
       phaCol(i) = i2pha
       if (i2off /= 0) koff = koff + 1
       if (i2pha > 0 .and. i2pha <= maxspec) &
                           i4buf(iy,i2pha) = i4buf(iy,i2pha) + 1
     end if 
   end do
     
   write(errstr,'(i9,a,i9,a)') koff, ' of', nevent, ' events selected for offset correction'
   call message(AppMsg,Sparse)
   
! spatial smoothing/filtering of the spectra:
   do k = 1, maxspec
     do j = 1, 19
       i4buff(j,k) = i4buf(j,k)
     end do
     do j = 20, nydim
       ky = 0
       do jy = max(j-10,20), min(j+10,nydim)
         ky = ky + 1
         i4med(ky) = i4buf(jy,k)
       end do
       call iindexx(ky,i4med,indx)
       i4buff(j,k) = i4med(indx((ky+1)/2))
     end do
   end do

   spcfrc=0.
   do i = minspec, maxspec
     do ky = 1, nydim
       h1 = noisemap(0,i-minspec,ky-1)
       h2 = real(max(i4buff(ky,i),1)) / expo

       spcfrc(ky,i) = amin1(amax1(h1/h2,0.0),noiseparam(0))
       spcfrc(ky,i) = spcfrc(ky,i) * noiseparam(iccd)
        
!       if (iccd.eq.4) then
!         spcfrc(ky,i) = spcfrc(ky,i) * 1.04
!       else if (iccd.eq.5) then
!         spcfrc(ky,i) = spcfrc(ky,i) * 1.07
!       else if (iccd.eq.6) then
!         spcfrc(ky,i) = spcfrc(ky,i) * 0.95
!       else if (iccd.le.3) then
!         spcfrc(ky,i) = spcfrc(ky,i) * 0.98
!       endif

     enddo
   enddo

!   open(1,file='spcfrc.dat',status='unknown')
!   do j = 1, nydim
!     write(1,'(40f5.2)') (spcfrc(j,k),k=20,maxspec)
!   end do
!   close(1)

   idum=-1
   ksel=0
   do i=0,nevent-1
     i2pha = phaCol(i)
     if (i2pha > maxspec) then
       noiseCol(i) = 0
     else
       iy = rawYCol(i)
       if (ran2(idum).gt.dble(spcfrc(iy,i2pha))) then
         noiseCol(i) = 0
       else
         noiseCol(i) = 1
         ksel = ksel + 1
       endif
     endif
   end do

   write(errstr,'(i9,a,f6.2,a)') &
                ksel,' events flagged  (', &
                100.0*real(ksel)/real(nevent),' % )'
   call message(AppMsg,Sparse)
   write(errstr,*) 'exposure: ',expo
   call message(AppMsg,Sparse)

   call release(offCor)
   call release(noise)

   else

!---------------------------------------------------
!---------------------------------------------------
!-  Vadims flare code start
!---------------------------------------------------
!---------------------------------------------------
   if (index(submode,"iming")==0.and.index(submode,"TIMING")==0) &
     call warning("WrongSubMode","Wrong observing mode: should be Timing Mode")

! append FLARE column
   if (hasColumn(events,"FLARE")) then
     flare = column(events,"FLARE",MODIFY)  
   else
     flare = addColumn (events, "FLARE", Integer16, "")
   end if   
   flareCol => int16Data(flare)
   flareCol = 0
    
    
     ndlc=2000000

!  withsoftflarescreening -- optional,type=bool,default=no; turns on flare screening code
!  softflarethreshold1 -- optional,type=real,default=8; threshold for flare screening
!  softflarethreshold2 -- optional,type=real,default=1; threshold for flare screening
!  softflaresmooth -- optional,type=string,default=BOX; smoothing method for flare screening (on of BOX, GAUSS, or FLARE)
!  softflareenergyrange -- optional,type=list<real>,default='0 60'; energy range for flare screening
!  softflaresmoothparams -- optional,type=list<real>,default='5 0 0'; smoothing parameters

   write(errstr,*) 'Soft flare screening started'
   call message(AppMsg,Sparse)


!-- energy range parameter
   npharangeelements=parametercount("softflareenergyrange")
   do i=0,npharangeelements-1
     thrpha(i) = intParameter("softflareenergyrange",i)
   end do
   if (npharangeelements == 1) then 
     thrpha(1)= thrpha(0)
     thrpha(0)=0
   endif

!-- smoothing  parameters
!--   nsmoothelements=parametercount("softflaresmoothparams")
!--   do i=0,nsmoothelements-1
     smo(0) = realParameter("softflaresmoothparams",0)
!--  end do
   
!-- cutoff threshhold for boxcar smoothed lightcurve 
     thrsig = realParameter("softflarethreshold1")

!--  make low energy lightcurve below pha threshhold
!--
     tbinsize = 0.1d0
     lc=0d0
     t0=timeCol(0)
     kmax=0
     do i=0,nevent-1 
        if (phaCol(i) .le. thrpha(1) .and. phaCol(i) .ge. thrpha(0)) then
	  k=nint((timeCol(i)-t0)/tbinsize)+1
	  if (k .lt. ndlc) then 
	    lc(k)=lc(k)+1
          else 
	    call error("Lightcurve has more bins than intrinsically assigned 2000000 for 0.1sec bins Hex # F8A67B secret file 21b SAS")
	  endif
	endif
	kmax=max(kmax,k)
     enddo
        
!--  smooth light curve with boxcar
!--
   
     sbox=nint(smo(0))
     do i = 1, kmax
       k=0
       ss=0
       do j = max(i-sbox,0), min(i+sbox,kmax)
         k=k+1
	 ss=ss+lc(j)
         enddo
       lcs(i) = ss/real(k)
      enddo

!--  set flare flag 
!--
     flareCol=0
     flagged=0
     do i=0, nevent-1
       if (phaCol(i) .le. thrpha(1) .and. phaCol(i) .ge. thrpha(0)) then
          k=nint((timeCol(i)-t0)/tbinsize)+1
          if (lcs(k) .gt. thrsig) then
            flareCol(i)=1
            flagged=flagged+1
          end if
       endif
     enddo
    

   write(errstr,*) 'Soft flare screening completed ', flagged, ' of', nevent ,' events flagged'
   call message(AppMsg,Sparse)

   call release(flare)

!---------------------------------------------------
!-  Vadims flare code end
!---------------------------------------------------
   end if

 
   call release(time)
   call release(ccdnr)
   call release(rawX)
   call release(rawY)
   call release(pha)
   call release(events)
   call release(infile)

!   if (booleanParameter("withcontrolimages")) then 

! write control images:
!     open(1,file='spcfrc.dat',status='unknown')
!     do j = 1, nydim
!       write(1,'(40f5.2)') (spcfrc(j,k),k=20,maxspec)
!     end do
!     close(1)

!     npix(1) = nxdim
!     npix(2) = nydim

!     write(string,'(a4,i2.2,a4)') 'mask',iccd,'.img'         
!     outfile=dataset(string,Create)
!     ima=addArray(outfile,"PRIMARY",Integer16,npix)
!     i2map => Int16Array2Data(ima)
      
!     do ix=0,nxdim-1
!       do iy=0,nydim-1
!         if (mask(ix+1,iy+1)) then
!           i2map(ix,iy)=1
!         else
!           i2map(ix,iy)=0
!         end if
!       end do
!     end do          

!     call release(ima)
!     call release(outfile)   

!     write(string,'(a6,i2.2,a4)') 'adu20_',iccd,'.img'         
!     outfile=dataset(string,Create)
!     ima=addArray(outfile,"PRIMARY",Integer16,npix)
!     i2map => Int16Array2Data(ima)
      
!     do ix=0,nxdim-1
!       do iy=0,nydim-1
!         i2map(ix,iy)=i4pmap(ix+1,iy+1)
!       end do
!     end do          

!     call release(ima)
!     call release(outfile)   

!     write(string,'(a6,i2.2,a4)') 'adu25_',iccd,'.img'         
!     outfile=dataset(string,Create)
!     ima=addArray(outfile,"PRIMARY",Integer16,npix)
!     i2map => Int16Array2Data(ima)
      
!     do ix=0,nxdim-1
!       do iy=0,nydim-1
!         i2map(ix,iy)=i30map(ix+1,iy+1)
!       end do
!     end do          

!     call release(ima)
!     call release(outfile)   

!     write(string,'(a6,i2.2,a4)') 'offcor',iccd,'.img'         
!     outfile=dataset(string,Create)
!     ima=addArray(outfile,"PRIMARY",Integer16,npix)
!     i2map => Int16Array2Data(ima)
      
!     do ix=0,nxdim-1
!       do iy=0,nydim-1
!         i2map(ix,iy)=i2pmap(ix+1,iy+1)
!       end do
!     end do          

!     call release(ima)
!     call release(outfile)   

!     write(string,'(a6,i2.2,a4)') 'spcmap',iccd,'.img'         
!     outfile=dataset(string,Create)
!     ima=addArray(outfile,"PRIMARY",Real32,npix)
!     r4map => real32Array2Data(ima)

!     do iy = 1, nydim
!       h = 0.0
!       do i = 22, 30
!         h = h + real(i4spc(iccd,iy,i))
!       end do
!       h = h / texp
!       do ix = 0, 63
!         r4map(ix,iy-1) = h
!       end do
!     end do

!     call release(ima)
!     call release(outfile)

!     write(string,'(a6,i2.2,a4)') 'bufmap',iccd,'.img'         
!     outfile=dataset(string,Create)
!     ima=addArray(outfile,"PRIMARY",Real32,npix)
!     r4map => real32Array2Data(ima)

!     do iy = 1, nydim
!       h = 0.0
!       do i = 22, 30
!         h = h + real(i4buf(iy,i))
!       end do
!       h = h / texp
!       do ix = 0, 63
!         r4map(ix,iy-1) = h
!       end do
!     end do

!     call release(ima)
!     call release(outfile)

!     write(string,'(a6,i2.2,a4)') 'frcmap',iccd,'.img'         
!     outfile=dataset(string,Create)
!     ima=addArray(outfile,"PRIMARY",Real32,npix)
!     r4map => real32Array2Data(ima)

!     do iy = 1, nydim
!       h = 0.0
!       do i = 22, 30
!         h = h + real(spcfrc(iy,i))
!       end do
!       do ix = 0, 63
!         r4map(ix,iy-1) = h
!       end do
!     end do

!     call release(ima)
!     call release(outfile)

!   end if

   end subroutine epreject
 
      FUNCTION ran2(idum)
      INTEGER idum,IM1,IM2,IMM1,IA1,IA2,IQ1,IQ2,IR1,IR2,NTAB,NDIV
      DOUBLE PRECISION ran2,AM,EPS,RNMX
      PARAMETER (IM1=2147483563,IM2=2147483399,AM=1.d0/IM1,IMM1=IM1-1, &
      IA1=40014,IA2=40692,IQ1=53668,IQ2=52774,IR1=12211,IR2=3791, &
      NTAB=32,NDIV=1+IMM1/NTAB,EPS=1.2d-7,RNMX=1.d0-EPS)
      INTEGER idum2,j,k,iv(NTAB),iy
      SAVE iv,iy,idum2
      DATA idum2/123456789/, iv/NTAB*0/, iy/0/
      if (idum.le.0) then
        idum=max(-idum,1)
        idum2=idum
        do j=NTAB+8,1,-1
          k=idum/IQ1
          idum=IA1*(idum-k*IQ1)-k*IR1
          if (idum.lt.0) idum=idum+IM1
          if (j.le.NTAB) iv(j)=idum
        end do
        iy=iv(1)
      endif
      k=idum/IQ1
      idum=IA1*(idum-k*IQ1)-k*IR1
      if (idum.lt.0) idum=idum+IM1
      k=idum2/IQ2
      idum2=IA2*(idum2-k*IQ2)-k*IR2
      if (idum2.lt.0) idum2=idum2+IM2
      j=1+iy/NDIV
      iy=iv(j)-idum2
      iv(j)=idum
      if(iy.lt.1)iy=iy+IMM1
      ran2=min(AM*iy,RNMX)
      return
      END function ran2

end module epreject_mod

